/*
 * app_base.hpp
 *
 * Copyright (c) Stephen Thompson, 2006.
 * Licenced for non-commercial use only. See LICENCE.txt for details.
 *
 * Shared code between 2D and 3D app classes.
 *
 */

#ifndef APP_BASE_HPP
#define APP_BASE_HPP

#include <set>
using namespace std;

// base class for exceptions.
// NB this does not copy the pointer
class AppError : public exception {
public:
	AppError(const char *m) : msg(m) { }
	virtual const char * what() const throw() { return msg; }
private:
	const char *msg;
};


//
// Key codes sent to AppObservers
// note normal character keys (on main keyboard) generate the normal ascii codes;
// these will be the unshifted values, but letters are sent as capitals.
// numeric keypad keys generate separate key codes.
// TODO Lock keys not yet supported
//
enum Keycode {
	KC_BACKSPACE = 8,
	KC_TAB = 9,
	KC_ESCAPE = 27,
	KC_RETURN = 13,   // main keyboard return key
	KC_F1 = 256, KC_F2, KC_F3, KC_F4, KC_F5, KC_F6,
	KC_F7, KC_F8, KC_F9, KC_F10, KC_F11, KC_F12,
	KC_LEFT, KC_RIGHT, KC_UP, KC_DOWN,
	KC_INSERT, KC_DELETE, KC_HOME, KC_END, KC_PAGE_UP, KC_PAGE_DOWN,
	KC_NUM_0, KC_NUM_1, KC_NUM_2, KC_NUM_3, KC_NUM_4,
	KC_NUM_5, KC_NUM_6, KC_NUM_7, KC_NUM_8, KC_NUM_9,
	KC_NUM_SLASH, KC_NUM_STAR, KC_NUM_MINUS, KC_NUM_PLUS, KC_NUM_ENTER, KC_NUM_DOT,
	KC_PRINT_SCREEN, KC_PAUSE,
	KC_LEFT_SHIFT, KC_RIGHT_SHIFT, KC_LEFT_CONTROL, KC_RIGHT_CONTROL,
	KC_LEFT_ALT, KC_RIGHT_ALT,
	KC_LEFT_WINDOWS, KC_RIGHT_WINDOWS, KC_MENU,
	KC_CAPS_LOCK, KC_NUM_LOCK, KC_SCROLL_LOCK
};


//
// AppObserver : base for classes wanting to respond to events, including render, update,
// activation/deactivation, keypresses
// For keypresses, mulitple key() events will be sent if the key auto-repeats, but only one
// keyUp() event is sent.
//

class AppObserver {
public:
	virtual ~AppObserver() { }
	virtual void render() { }
	virtual void update() { }
	virtual void onActivate() { }
	virtual void onDeactivate() { }
	virtual void key(Keycode key) { }
	virtual void keyUp(Keycode key) { }
	// Note: Mouse coords are relative to top-left of screen.
	virtual void mouseDown(int x, int y, bool rmb) { }  // LMB or RMB pressed.
	virtual void mouseUp(int x, int y, bool rmb) { }    // LMB or RMB released.
	virtual void mouseMove(int x, int y) { }            // Mouse pointer moved.
};


//
// Base class for apps
//

class AppBase {
public:

	//
	// Registration of observers
	//
	// Note: During AppObserver callbacks, it's ok for an observer to
	// remove itself, it's ok for an observer to add new observers,
	// but an observer should not attempt to remove other observers as
	// this can cause crashed. (This should really be fixed somehow --
	// maybe take a copy of the observer set and iterate over that?)

	void addObserver(AppObserver *ob) { observers.insert(ob); }
	void rmObserver(AppObserver *ob) { observers.erase(ob); }

	
	//
	// Activation: app is deactivated when window becomes hidden (eg minimized) and
	// reactivated when it becomes visible again. 
	// (Subclass should call activate() and deactivate() when these events occur.)
	// Client can also manually call activate() and deactivate().
	//

	void activate()
		{
			active = true;
			onActivate();
			resetUpdateIntervals();
			for (set<AppObserver*>::const_iterator it = observers.begin();
			it != observers.end(); ++it) {
				(*it)->onActivate();
			}
		}
	void deactivate()
		{
			active = false;
			onDeactivate();
			resetUpdateIntervals();
			for (set<AppObserver*>::const_iterator it = observers.begin();
			it != observers.end(); ++it) {
				(*it)->onDeactivate();
			}
		}
	bool isActive()
		{ return active; }

	
	//
	// Rendering:
	// When inactive: render() is called only when window needs repainting.
	// When active and render_fast is true: render() is called once per frame.
	//      (This generally means "as fast as possible". XXX perhaps should set 
	//      up more detailed mechanism to control this. Eg could have
	//      a render_interval instead of render_fast flag, 0 meaning "as fast as
	//      possible", positive specifying a time between renders, negative meaning
	//      "don't render at all" -- just like updates.)
	// When active and render_fast is false: render() is called after every
	// 		call to update() (unless updates are coming too fast; some renders
	// 		may be dropped in that case). NB if updates are compeletely disabled,
	// 		then render() will be called only after a system "repaint" event is
	// 		received.
	//
	// Updating:
	// update() is called periodically, according to the update intervals.
	// zero:     update once per frame.
	// positive: gives update interval in milliseconds (allows updates to be at a
	//             different rate to graphics frames).
	// negative: don't call update at all.
	// There are two intervals, one for when app is active and one for when inactive.
	//
	// Defaults:	update interval active = 0 (each frame),
	// 				update interval inactive = -1 (disabled),
	// 				render fast = false (render after updates only).
	//
	// Note it's up to subclasses to implement the above. The update() and
	// render() functions are defined below and call the relevant operations on
	// all registered observers.
	//

	void setUpdateIntervalActive(int milliseconds)
		{ update_a = milliseconds; resetUpdateIntervals(); }
	void setUpdateIntervalInactive(int milliseconds)
		{ update_i = milliseconds; resetUpdateIntervals(); }
	void setRenderFast(bool rf) { render_fast = rf; }

	int getUpdateIntervalActive() { return update_a; }
	int getUpdateIntervalInactive() { return update_i; }
	int getCurrentUpdateInterval()
		{ return isActive() ? getUpdateIntervalActive() : getUpdateIntervalInactive(); }
	bool getRenderFast() { return render_fast; }
	bool wantRenderAfterUpdate() { return !render_fast && active; }
	bool wantRenderEachFrame() { return render_fast && active; }
	bool wantRenderOnRepaint() { if (!active) return true; if (render_fast) return false; return update_a == -1; }


	//
	// show mouse cursor when full screen? defaults to false
	// XXX not sure if this is useful or not. at least not until mouse input method is worked
	// out.
	//
	bool showCursorFullScreen() { return curs_full; }
	void setShowCursorFullScreen(bool cf) { curs_full = cf; }


	//
	// Initialization: Subclasses should have a ctor to initialize the api and a dtor to
	// do any necessary cleanup and close any open display.
	// - assumes active to begin with (onActive is _not_ called initially)
	// - we do _not_ call resetUpdateIntervals initially
	//
	AppBase() : update_a(0), update_i(-1), render_fast(false), active(true),
		curs_full(false) { }
	virtual ~AppBase() { }


	//
	// Query functions for displays
	//

	// XXX display width/height could change spontaneously eg if user resizes a window.
	// Perhaps we should have an event (in AppObserver) for this.
	virtual int getDisplayWidth() = 0;
	virtual int getDisplayHeight() = 0;
	virtual bool isFullScreen() = 0;
	virtual bool isWindowed() = 0;
	bool displayOpen() { return isFullScreen() || isWindowed(); }

	// Does the app/display have input focus?
	// (eg could use this to deactivate if we find we've lost window focus.)
	virtual bool hasFocus() = 0;

	
	//
	// Main Loop
	//

	// Enter the main loop. Does not return until user closes window, or exitMainLoop called.
	virtual void enterMainLoop() = 0;

	// Tells the App to exit the main loop as soon as possible. (undefined effect if called
	// outside main loop.)
	virtual void exitMainLoop() = 0;


	//
	// Timer
	//

	// get number of ms elapsed since some (arbitrary) reference time
	virtual int getMilliseconds() = 0;

	
protected:
	//
	// this is called when any update interval changes, or when the app activates or
	// deactivates. intention is for things like timers to be reset.
	//
	virtual void resetUpdateIntervals() = 0;

	//
	// hooks called on activation/deactivation - but note that this class already deals with
	// active flag (returned by isActive) and calls resetUpdateIntervals to reset any timers.
	//
	virtual void onActivate() { }
	virtual void onDeactivate() { }

	//
	// sending events to observers: (note activate/deactivate are done separately)
	// These should be called by subclasses when events occur
	//
	void render()
		{
			set<AppObserver*>::const_iterator it = observers.begin();
			while (it != observers.end()) {
				AppObserver *p = *it++;
				p->render();
			}
		}
	void update()
		{
			set<AppObserver*>::const_iterator it = observers.begin();
			while (it != observers.end()) {
				AppObserver *p = *it++;
				p->update();
			}
		}
	void key(Keycode kc)
		{
			set<AppObserver*>::const_iterator it = observers.begin();
			while (it != observers.end()) {
				AppObserver *p = *it++;
				p->key(kc);
			}
		}
	void keyUp(Keycode kc)
		{
			set<AppObserver*>::const_iterator it = observers.begin();
			while (it != observers.end()) {
				AppObserver *p = *it++;
				p->keyUp(kc);
			}
		}
	void mouseDown(int x, int y, bool rmb) {
		set<AppObserver*>::const_iterator it = observers.begin();
		while (it != observers.end()) {
			AppObserver *p = *it++;
			p->mouseDown(x, y, rmb);
		}
	}
	void mouseUp(int x, int y, bool rmb) {
		set<AppObserver*>::const_iterator it = observers.begin();
		while (it != observers.end()) {
			AppObserver *p = *it++;
			p->mouseUp(x, y, rmb);
		}		
	}
	void mouseMove(int x, int y) {
		set<AppObserver*>::const_iterator it = observers.begin();
		while (it != observers.end()) {
			AppObserver *p = *it++;
			p->mouseMove(x,y);
		}
	}
	
private:
	set<AppObserver*> observers;
	int update_a, update_i;
	bool render_fast;
	bool active;
	bool curs_full;
};

#endif		// APP_BASE_HPP
