/*
 * sdl_app2d.cpp
 *
 * Copyright (c) Stephen Thompson, 2006.
 * Licenced for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "sdl_app2d.hpp"

#include <cassert>
#include <cctype>
using namespace std;


//
// Constructor, destructor
//

SdlApp2D::SdlApp2D()
: windowed(false), has_focus(false), timer_set(false), render_pending(false)
{
	int retval = SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER | SDL_INIT_NOPARACHUTE);
	if (retval != 0) throw AppError("SDL_Init failed");

	initKeyTable();
}

SdlApp2D::~SdlApp2D()
{
	SDL_Quit();
}



//
// Full screen displays
//

int SdlApp2D::findDisplayModes()
{
	display_modes.clear();

	SDL_Rect **modes = SDL_ListModes(0, SDL_FULLSCREEN | SDL_HWSURFACE | SDL_DOUBLEBUF);

	if (modes==0) return 0;
	if (modes==(SDL_Rect**)(-1)) {
		throw AppError("-1 return from SDL_ListModes not yet supported");
	}

	while (*modes) {
		DisplayMode dm;
		dm.width = (*modes)->w;
		dm.height = (*modes)->h;
		dm.colour_depth = SDL_GetVideoInfo()->vfmt->BitsPerPixel;
		display_modes.push_back(dm);
		++modes;
	}
	return display_modes.size();
}

const DisplayMode & SdlApp2D::getDisplayMode(int n)
{
	if (n < 0 || n >= display_modes.size()) {
		throw AppError("mode number out of range");
	} else {
		return display_modes[n];
	}
}

void SdlApp2D::openFullScreen(int mode_num, const char *title)
{
	if (mode_num < 0 || mode_num > display_modes.size()) throw AppError("mode number out of range");
	SDL_SetVideoMode(display_modes[mode_num].width,
					 display_modes[mode_num].height,
					 display_modes[mode_num].colour_depth,
					 SDL_FULLSCREEN | SDL_HWSURFACE | SDL_ANYFORMAT | SDL_DOUBLEBUF);
	windowed = false;
	has_focus = true;
	SDL_WM_SetCaption(title, title);
}



//
// Windowed displays
//

void SdlApp2D::openWindowed(int x, int y, int width, int height,
							const char *title)
{
	// x,y ignored since sdl does not expose an api to set these values.
	SDL_SetVideoMode(width, height, 0,
		SDL_ANYFORMAT | SDL_HWSURFACE | SDL_DOUBLEBUF | SDL_RESIZABLE);

	// set windowed flag etc
	windowed = true;
	has_focus = true;

	SDL_WM_SetCaption(title, title);
}


//
// Event handler
//
void SdlApp2D::handleEvent(const SDL_Event &evt)
{
	switch (evt.type) {
	case SDL_VIDEOEXPOSE:
		// Must redraw the screen (if app is inactive)
		if (wantRenderOnRepaint()) {
			render_pending = true;
		}
		break;

	case SDL_ACTIVEEVENT:
		switch (evt.active.state) {
		case SDL_APPINPUTFOCUS:
			// input focus gained or lost
			has_focus = evt.active.gain;
			break;
		case SDL_APPACTIVE:
			// app has been minimized or maximized.
			if (evt.active.gain) activate();
			else deactivate();
			break;
		}
		break;

	case SDL_VIDEORESIZE:
		// call SDL_SetVideoMode to select the new size.
		// --This only works for windowed mode, but fullscreen can't resize anyway.
		assert(windowed);
		SDL_SetVideoMode(evt.resize.w, evt.resize.h, 0,
			SDL_ANYFORMAT | SDL_HWSURFACE | SDL_DOUBLEBUF | SDL_RESIZABLE);
		break;

	case SDL_KEYDOWN:
		keyDownEvent(evt.key);
		break;

	case SDL_KEYUP:
		keyUpEvent(evt.key);
		break;

	case SDL_MOUSEBUTTONDOWN:
		if (evt.button.button == SDL_BUTTON_LEFT || evt.button.button == SDL_BUTTON_RIGHT) {
			mouseDown(evt.button.x, evt.button.y, evt.button.button == SDL_BUTTON_RIGHT);
		}
		break;

	case SDL_MOUSEBUTTONUP:
		if (evt.button.button == SDL_BUTTON_LEFT || evt.button.button == SDL_BUTTON_RIGHT) {
			mouseUp(evt.button.x, evt.button.y, evt.button.button == SDL_BUTTON_RIGHT);
		}
		break;

	case SDL_MOUSEMOTION:
		mouseMove(evt.motion.x, evt.motion.y);
		break;
		
	case SDL_USEREVENT:
		assert(evt.user.code == 1);	// we only have timer events (code 1) ...
		// Timer event - it is time to call update()
		update();
		if (wantRenderAfterUpdate()) {
			// tell main loop to render a frame, as soon as it has time.
			render_pending = true;
		}
		break;
		
	}
}

void SdlApp2D::keyDownEvent(const SDL_KeyboardEvent &kbd)
{
	if (keytable.find(kbd.keysym.sym) != keytable.end()) {
		key(keytable[kbd.keysym.sym]);
	} else {
		key(Keycode(toupper(kbd.keysym.sym)));
	}	
}

void SdlApp2D::keyUpEvent(const SDL_KeyboardEvent &kbd)
{
	if (keytable.find(kbd.keysym.sym) != keytable.end()) {
		keyUp(keytable[kbd.keysym.sym]);
	} else {
		keyUp(Keycode(toupper(kbd.keysym.sym)));
	}
}



//
// key table implementation
//

void SdlApp2D::initKeyTable()
{
	keytable[SDLK_BACKSPACE] = KC_BACKSPACE;
	keytable[SDLK_TAB] = KC_TAB;
	keytable[SDLK_ESCAPE] = KC_ESCAPE;
	keytable[SDLK_RETURN] = KC_RETURN;
	keytable[SDLK_KP_ENTER] = KC_NUM_ENTER; 
	keytable[SDLK_F1] = KC_F1;
	keytable[SDLK_F2] = KC_F2;
	keytable[SDLK_F3] = KC_F3;
	keytable[SDLK_F4] = KC_F4;
	keytable[SDLK_F5] = KC_F5;
	keytable[SDLK_F6] = KC_F6;
	keytable[SDLK_F7] = KC_F7;
	keytable[SDLK_F8] = KC_F8;
	keytable[SDLK_F9] = KC_F9;
	keytable[SDLK_F10] = KC_F10;
	keytable[SDLK_F11] = KC_F11;
	keytable[SDLK_F12] = KC_F12;
	keytable[SDLK_LEFT] = KC_LEFT;
	keytable[SDLK_RIGHT] = KC_RIGHT;
	keytable[SDLK_UP] = KC_UP;
	keytable[SDLK_DOWN] = KC_DOWN;
	keytable[SDLK_INSERT] = KC_INSERT;
	keytable[SDLK_DELETE] = KC_DELETE;
	keytable[SDLK_HOME] = KC_HOME;
	keytable[SDLK_END] = KC_END;
	keytable[SDLK_PAGEUP] = KC_PAGE_UP;
	keytable[SDLK_PAGEDOWN] = KC_PAGE_DOWN;
	keytable[SDLK_KP0] = KC_NUM_0;
	keytable[SDLK_KP1] = KC_NUM_1;
	keytable[SDLK_KP2] = KC_NUM_2;
	keytable[SDLK_KP3] = KC_NUM_3;
	keytable[SDLK_KP4] = KC_NUM_4;
	keytable[SDLK_KP5] = KC_NUM_5;
	keytable[SDLK_KP6] = KC_NUM_6;
	keytable[SDLK_KP7] = KC_NUM_7;
	keytable[SDLK_KP8] = KC_NUM_8;
	keytable[SDLK_KP9] = KC_NUM_9;
	keytable[SDLK_KP_DIVIDE] = KC_NUM_SLASH;
	keytable[SDLK_KP_MULTIPLY] = KC_NUM_STAR;
	keytable[SDLK_KP_MINUS] = KC_NUM_MINUS;
	keytable[SDLK_KP_PLUS] = KC_NUM_PLUS;
	keytable[SDLK_KP_PERIOD] = KC_NUM_DOT;
	keytable[SDLK_PRINT] = KC_PRINT_SCREEN;
	keytable[SDLK_PAUSE] = KC_PAUSE;
	keytable[SDLK_LSHIFT] = KC_LEFT_SHIFT;
	keytable[SDLK_RSHIFT] = KC_RIGHT_SHIFT;
	keytable[SDLK_LCTRL] = KC_LEFT_CONTROL;
	keytable[SDLK_RCTRL] = KC_RIGHT_CONTROL;
	keytable[SDLK_LALT] = KC_LEFT_ALT;
	keytable[SDLK_RALT] = KC_RIGHT_ALT;
	keytable[SDLK_LSUPER] = KC_LEFT_WINDOWS;
	keytable[SDLK_RSUPER] = KC_RIGHT_WINDOWS;
	keytable[SDLK_MENU] = KC_MENU;
	keytable[SDLK_CAPSLOCK] = KC_CAPS_LOCK;
	keytable[SDLK_NUMLOCK] = KC_NUM_LOCK;
	keytable[SDLK_SCROLLOCK] = KC_SCROLL_LOCK;
}





//
// Set timers up so that they post a user message with ID code 1, and set
// the next timer interval to be the same as last time.
//

Uint32 MyTimerCallback(Uint32 interval, void *)
{
	SDL_Event evt;
	evt.type = SDL_USEREVENT;
	evt.user.code = 1;
	evt.user.data1 = evt.user.data2 = 0;
	SDL_PushEvent(&evt);
	return interval;
}

void SdlApp2D::resetUpdateIntervals()
{
	if (timer_set) {
		SDL_RemoveTimer(timer_id);
		timer_set = false;
	}

	if (getCurrentUpdateInterval() > 0) {
		timer_set = true;
		// round interval up to next 10ms due to granularity of SDL timers (!!!)
		timer_id = SDL_AddTimer((getCurrentUpdateInterval()/10)*10,
			 MyTimerCallback, 0);
	}
}



//
// Main Loop
//

void SdlApp2D::enterMainLoop()
{
	// kickstart the timers
	resetUpdateIntervals();

	// Main message loop
	SDL_Event evt;

	while (1) {
		// Check for events.
		// If there is stuff to do imminently then we use PollEvent.
		// Otherwise we use waitEvent to free up some cpu time.

		// (TODO) this might be problematic if someone changes the settings mid game... may
		// have to send some sort of user event to wake the main loop up after certain
		// changes?

		bool stuff_to_do = (getCurrentUpdateInterval() == 0 || wantRenderEachFrame() || render_pending);
		bool got_event = true;
		if (stuff_to_do) got_event = SDL_PollEvent(&evt);
		else SDL_WaitEvent(&evt);

		if (got_event) {
			// Handle the event
			handleEvent(evt);
			if (evt.type == SDL_QUIT) break;

		} else {
			// Update if necessary
			if (getCurrentUpdateInterval() == 0) {
				update();
				if (wantRenderAfterUpdate()) render_pending = true;
			}

			// Render if necessary
			if (wantRenderEachFrame() || render_pending) {
				render();
				render_pending = false;
			}
		}

		// TODO:
		// Adding the following really slows things down. That concerns me (esp as
		// this is basically what we do on the PPC)... looks like we
		// need to overhaul the app code!
		//		if (render_pending) {
		//			render();
		//			render_pending = false;
		//		}
	}
}

void SdlApp2D::exitMainLoop()
{
	SDL_Event evt;
	evt.type = SDL_QUIT;
	SDL_PushEvent(&evt);
}


//
// Query functions.
//

int SdlApp2D::getDisplayWidth()
{
	return SDL_GetVideoSurface()->w;
}

int SdlApp2D::getDisplayHeight()
{
	return SDL_GetVideoSurface()->h;
}

bool SdlApp2D::isFullScreen()
{
	return bool(SDL_GetVideoSurface()->flags | SDL_FULLSCREEN);
}

bool SdlApp2D::isWindowed()
{
	return !bool(SDL_GetVideoSurface()->flags | SDL_FULLSCREEN);
}

bool SdlApp2D::hasFocus()
{
	return has_focus;
}

//
// Time
//

int SdlApp2D::getMilliseconds()
{
	return SDL_GetTicks();
}
