/*
 * creature.hpp
 *
 * Copyright (c) Stephen Thompson, 2006.
 * Licenced for non-commercial use only. See LICENCE.txt for details.
 *
 * Base class for 'creatures' which currently includes knights,
 * vampire bats and zombies. (Derived from Entity.)
 *
 * This class implements combat including item-in-hand support,
 * swinging and parrying, a hitpoints system, and a 'stun' timer.
 *
 */

#ifndef CREATURE_HPP
#define CREATURE_HPP

#include "entity.hpp"

class ItemType;

class Creature : public Entity {
	friend class ImpactTask;
	friend class ThrowingTask;
public:
	// ctor, dtor
	Creature(int health, MapHeight height, const ItemType * item_in_hand,
			 const Anim * lower_anim, int base_speed);

	// various accessors
	int getHealth() const { return health; }
	int getMaxHealth() const { return max_health; }
	virtual MapHeight getHeight() const { return height; }

	// stunned: means the creature can't do anything (because
	// paralyzed, just been whacked by a hammer, in the middle of
	// attacking or moving, etc.) If stunned is false then the
	// creature is not moving and ready to receive orders.
	bool stunned() const;

	// stunnedUntil: If we know when the creature will become
	// unstunned (or we know a lower bound for that time), then
	// 'known' will be set to true, and 'stunned_until' will be set to
	// (a lower bound for) that time. (It is guaranteed that
	// 'stunned_until' will be greater than the current gvt.) If we
	// have no idea when we will become unstunned, then 'known' will
	// be set to false (and 'stunned_until' will be left unchanged).
	void stunnedUntil(bool &known, int &stunned_until) const;

	// stun the creature until given time
	void stunUntil(int new_stun_time);

	// motionActionPossible: means the creature is currently "stunned", but
	// only because he is moving (and no other reason); this means
	// that actions such as "cancel move" and "attack while moving"
	// are possible.
	bool motionActionPossible() const;
	
	// item in hand
	const ItemType * getItemInHand() const { return item_in_hand; }
	virtual void setItemInHand(const ItemType * i)  // pass 0 to mean 'no item in hand'
		{ doSetItemInHand(i); }

	//
	// combat / hitpoints stuff 
	//
	// damage(): this is virtual, and can be overridden although
	// you'll probably have to call Creature::damage as part of your
	// implementation. Eg Knight does this to filter out damage for
	// invulnerabele knights, and to update the dungeonview.
	//
	// addToHealth(): likewise.
	//
	// poison(): poison damage. Default implementation kills the
	// creature outright. Knight overrides in order to implement
	// poison immunity and invulnerability.
	//
	// impactVeto(): This is called just before a weapon impacts
	// against this creature. If it returns TRUE then the impact is
	// cancelled ("vetoed"). 
	//

	bool canSwing() const;
	void swing();  // start a melee attack, setting anim and stun_time.

	bool canThrow(bool strict) const;
	void throww();  // start a throwing attack, setting anim and stun_time.
	
	virtual void damage(int amount, int stun_until = -1);
	virtual int bloodLevel() const { return 2; }  // 1=only blood when die 0=no blood at all
	virtual void addToHealth(int amount);
	virtual void poison();
	bool impactVeto(int gvt, const Creature &attacker);

	void parry();
	
	// strength and quickness: these are overridden by Knights.
	// onDeath: is called by "damage" just before removing the dead creature from the map.
	// Used by Knights, to drop items before leaving. Also onDeath is responsible for placing
	// the corpse.
	// -- onDeath can also be called directly (eg this is done by PitKill). However note that
	// this won't take invulnerability into account.
	enum DeathMode { PIT_MODE,  // items to be dropped one square back; no corpse to be placed
					 POISON_MODE,  // corpse to be placed but no blood
					 NORMAL_MODE,  // corpse to be placed, with blood
	};
	virtual void onDeath(DeathMode dmode) { }
	virtual bool hasStrength() const { return false; }
	virtual bool hasQuickness() const { return false; }

	int getBaseSpeed() const { return speed; }
	
protected:
	void doSetItemInHand(const ItemType *i);
	virtual bool isParalyzed() const { return false; } // if true, then makes the creature stunned (for an unknown time).

	// routines for throwing daggers, axes etc
	// in canDoThrow: strict==false means turn off certain checks, eg stunned() check.
	bool canDoThrow(bool strict) const;
	void doThrow(const ItemType &);
	virtual void throwAwayItem(const ItemType *);
	
private:
	void doWithdrawEvents();

private:
	MapHeight height;
	int health, max_health;
	int speed; // base speed, before quickness is applied
	int stun_until;  // 0 if not stunned. NB stun_until is guaranteed to be non-decreasing.
	const ItemType * item_in_hand;
	int impact_time;  // if nonzero, the creature is attacking and will hit at this time.
};

#endif
