/*
 * entity_map.cpp
 *
 * Copyright (c) Stephen Thompson, 2006.
 * Licenced for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "anim.hpp"
#include "cfg.hpp"
#include "entity_map.hpp"
#include "graphic_element.hpp"

int EntityMap::getFinalOffset(MotionType type)
{
	switch (type) {
	case MT_MOVE: return CfgInt("points_per_square");
	case MT_APPROACH: return CfgInt("approach_offset");
	case MT_WITHDRAW: return 0;
	case MT_REVERSE: return 0;
	default: ASSERT(0); return 0;
	}
}

void EntityMap::addEntity(int gvt, unsigned short int key, int x, int y, MapHeight h, MapDirection facing,
						  const Anim *anim, const Overlay *ovr, int af, int atz, bool ainvuln,
						  int cur_ofs, MotionType motion_type, int motion_time_remaining)
{
	Data d;
	d.x = x;
	d.y = y;
	d.height = h;
	d.anim = anim;
	d.ovr = ovr;
	d.af = af;
	d.atz = atz;
	d.ainvuln = ainvuln;
	d.facing = facing;
	d.approached = (motion_type == MT_NOT_MOVING && cur_ofs != 0);
	if (motion_type == MT_NOT_MOVING) {
		d.start_time = 0;
		d.finish_time = 0;
		d.tnt = 0;
	} else {
		d.start_time = gvt;
		d.finish_time = gvt + motion_time_remaining;
		d.tnt = motion_time_remaining;
		Command cmd;
		cmd.type = Command::MOVE;
		cmd.move_info.so = cur_ofs;
		cmd.move_info.type = motion_type;
		cmd.move_info.nt = motion_time_remaining;
		d.cmds.push_back(cmd);
	}
	entities.insert(make_pair(key, d));
}

void EntityMap::rmEntity(unsigned short int key)
{
	map<unsigned short int,Data>::iterator it = entities.find(key);
	if (it == entities.end()) return;

	entities.erase(it);
}

void EntityMap::moveEntity(int gvt, unsigned short int key,
						   MotionType motion_type, int motion_duration)
{
	if (motion_type == MT_NOT_MOVING) return;
	
	map<unsigned short int,Data>::iterator it = entities.find(key);
	if (it == entities.end()) return;
	update(gvt, it->second);  // Make sure any "reposition" or similar commands are popped
							  // from the top of the command queue.

	// new starting offset
	int new_so = it->second.approached? CfgInt("approach_offset") : 0;
	
	if (!it->second.cmds.empty()) {
		// A motion command is in progress.
		Command &cmd(it->second.cmds.front());
		ASSERT(cmd.type == Command::MOVE);
		ASSERT(cmd.move_info.nt != 0);
		ASSERT(it->second.start_time < it->second.finish_time);

		// calculate the offset of the entity right now (taking into
		// account the motion cmd that is currently in progress).
		int nt_so_far = (gvt - it->second.start_time) * it->second.tnt
			/ (it->second.finish_time - it->second.start_time);
		if (nt_so_far < 0) nt_so_far = 0;
		if (nt_so_far > cmd.move_info.nt) nt_so_far = cmd.move_info.nt;
		const int cur_ofs = cmd.move_info.so
			+ (getFinalOffset(cmd.move_info.type) - cmd.move_info.so)
			* nt_so_far / cmd.move_info.nt;
		
		if (motion_type == MT_REVERSE) {
			// Cancel the existing command.
			ASSERT(cmd.move_info.type == MT_MOVE);
			new_so = cur_ofs;
			it->second.tnt -= cmd.move_info.nt;
			it->second.cmds.pop_front();
		} else {
			// Update the existing command, by changing its start time
			// to current gvt, and updating its start offset to
			// current offset. This will prevent the current offset
			// from 'jumping' when we add the new command.
			cmd.move_info.so = cur_ofs;
			cmd.move_info.nt -= nt_so_far;
			it->second.tnt -= nt_so_far;
		}
	} else {
		ASSERT(it->second.tnt == 0);
	}

	it->second.start_time = gvt;
	
	// Add a new command to represent the move
	Command cmd;
	cmd.type = Command::MOVE;
	cmd.move_info.so = new_so;
	cmd.move_info.nt = motion_duration;
	cmd.move_info.type = motion_type;
	it->second.cmds.push_back(cmd);
	it->second.approached = (motion_type == MT_APPROACH);
	it->second.finish_time = gvt + motion_duration;
	it->second.tnt += cmd.move_info.nt;
}

void EntityMap::repositionEntity(unsigned short int key, int new_x, int new_y)
{
	map<unsigned short int,Data>::iterator it = entities.find(key);
	if (it == entities.end()) return;
	
	Command cmd;
	cmd.type = Command::REPOSITION;
	cmd.reposition_info.x = new_x;
	cmd.reposition_info.y = new_y;
	it->second.cmds.push_back(cmd);
	it->second.approached = false;
}

void EntityMap::setAnimData(unsigned short int key, const Anim *anim, const Overlay *ovr,
							int af, int atz, bool ainvuln)
{
	map<unsigned short int, Data>::iterator it = entities.find(key);
	if (it == entities.end()) return;
	Command cmd;
	cmd.type = Command::SET_ANIM;
	cmd.anim_info.anim = anim;
	cmd.anim_info.ovr = ovr;
	cmd.anim_info.af = af;
	cmd.anim_info.atz = atz;
	cmd.anim_info.ainvuln = ainvuln;
	it->second.cmds.push_back(cmd);
}

void EntityMap::setFacing(unsigned short int key, MapDirection f)
{
	map<unsigned short int,Data>::iterator it = entities.find(key);
	if (it == entities.end()) return;
	Command cmd;
	cmd.type = Command::SET_FACING;
	cmd.facing_info = f;
	it->second.cmds.push_back(cmd);
}

void EntityMap::update(int gvt, EntityMap::Data &ent)
{
	while (!ent.cmds.empty()) {
		const Command &cmd(ent.cmds.front());
			
		switch (cmd.type) {
		case Command::MOVE:
			{
				int dur = (ent.finish_time - ent.start_time);
				int fini_time = (cmd.move_info.nt * dur / ent.tnt)
					+ ent.start_time;
				if (gvt < fini_time) {
					// This command is still in progress
					return;
				} else {
					// This command has completed
					ent.tnt -= cmd.move_info.nt;
					// Next command starts when previous one finished:
					ent.start_time = fini_time;
					// If it was a "move" command, we have to update position, too:
					// (it's no good waiting for the server to send the reposition command,
					// because while we're waiting, the entity will have the wrong
					// pos/offset.)
					if (cmd.move_info.type == MT_MOVE) {
						switch (ent.facing) {
						case D_NORTH:
							++ent.y;
							break;
						case D_EAST:
							++ent.x;
							break;
						case D_SOUTH:
							--ent.y;
							break;
						case D_WEST:
							--ent.x;
							break;
						}
					}
				}
			}
			break;
			
		case Command::REPOSITION:
			ent.x = cmd.reposition_info.x;
			ent.y = cmd.reposition_info.y;
			break;
			
		case Command::SET_ANIM:
			ent.anim = cmd.anim_info.anim;
			ent.ovr = cmd.anim_info.ovr;
			ent.af = cmd.anim_info.af;
			ent.atz = cmd.anim_info.atz;
			ent.ainvuln = cmd.anim_info.ainvuln;
			break;
			
		case Command::SET_FACING:
			ent.facing = cmd.facing_info;
			break;
		}
		
		ent.cmds.pop_front();
	}
}	
	
	
void EntityMap::update(int gvt)
{
	for (map<unsigned short int,Data>::iterator it = entities.begin(); it != entities.end(); ++it) {
		update(gvt, it->second);
	}
}

void EntityMap::addGraphic(const Data &ent, int gvt, int bl_x, int bl_y, int entity_depth,
						   map<int, vector<GraphicElement> > &gfx_buffer,
						   int pixels_per_square, int points_per_square, int approach_offset)
{
	const Anim *anim = ent.anim;
	if (anim) {
		
		// Work out what graphics to draw, and where to draw them...
		const int frame = (gvt >= ent.atz && ent.atz > 0) ? 0 : ent.af;
		const Graphic *lower_graphic = anim->getGraphic(ent.facing, frame, ent.x, ent.y, gvt);
		const ColourChange *lower_cc = &(anim->getColourChange(ent.ainvuln));

		if (lower_graphic) {
			int ox, oy;
			int ofs;
			if (ent.cmds.empty()) {
				ofs = ent.approached? approach_offset : 0;
			} else {
				const Command &cmd(ent.cmds.front());
				ASSERT(cmd.type == Command::MOVE);
				const int fo = getFinalOffset(cmd.move_info.type);
				// following is done as two divisions, to prevent overflow in
				// cases where the entity moves very slowly...
				ofs = cmd.move_info.so +
					(fo - cmd.move_info.so) * ent.tnt / cmd.move_info.nt
					* (gvt - ent.start_time) / (ent.finish_time - ent.start_time);
			}
			switch (ent.facing) {
			case D_NORTH:
				ox = 0;
				oy = ofs;
				break;
			case D_EAST:
				ox = ofs;
				oy = 0;
				break;
			case D_SOUTH:
				ox = 0;
				oy = -ofs;
				break;
			case D_WEST:
				ox = -ofs;
				oy = 0;
				break;
			}
			
			const int d = entity_depth - 2*ent.height;				
			
			GraphicElement ge;
			ge.sx = bl_x
				+ ent.x * pixels_per_square
				+ ox * pixels_per_square / points_per_square;
			ge.sy = bl_y
				+ ent.y * pixels_per_square
				+ oy * pixels_per_square / points_per_square;
			ge.gr = lower_graphic;
			ge.cc = lower_cc;
			gfx_buffer[d].push_back(ge);
			
			const Overlay *ovr = ent.ovr;
			if (ovr) {
				int upper_ofsx, upper_ofsy;
				MapDirection mdir;
				bool show;
				GetOverlayOffset(ent.facing, frame, upper_ofsx, upper_ofsy, mdir, show);
				if (show) {
					ge.gr = ovr->getGraphic(mdir);
					ge.cc = 0;
					ge.sx += upper_ofsx;
					ge.sy += upper_ofsy;
					gfx_buffer[d-1].push_back(ge);
				}
			}
		}
	}
}	

void EntityMap::getEntityGfx(int gvt, int bl_x, int bl_y, int entity_depth,
							 map<int, vector<GraphicElement> > &gfx_buffer)
{
	const int pixels_per_square = CfgInt("pixels_per_square");
	const int points_per_square = CfgInt("points_per_square");
	const int approach_offset = CfgInt("approach_offset");
	
	update(gvt);
	
	for (map<unsigned short int,Data>::iterator it = entities.begin(); it != entities.end(); ++it) {
		addGraphic(it->second, gvt, bl_x, bl_y, entity_depth, gfx_buffer,
				   pixels_per_square, points_per_square, approach_offset);
	}
}

