/*
 * entity_map.hpp
 *
 * Copyright (c) Stephen Thompson, 2006.
 * Licenced for non-commercial use only. See LICENCE.txt for details.
 *
 * Holds Entities within a LocalDungeonView.
 *
 */

#ifndef ENTITY_MAP_HPP
#define ENTITY_MAP_HPP

#include "map_support.hpp"

#include <deque>
#include <list>
#include <map>
#include <vector>
using namespace std;

class Anim;
class GraphicElement;
class Overlay;

class EntityMap {
public:
	void addEntity(int gvt, unsigned short int key, int x, int y, MapHeight h, MapDirection facing,
				   const Anim *anim, const Overlay *ovr, int af, int atz, bool ainvuln,
				   int cur_ofs, MotionType motion_type, int motion_time_remaining);
	void rmEntity(unsigned short int key);
	void moveEntity(int gvt, unsigned short int key, MotionType motion_type, int motion_duration);
	void repositionEntity(unsigned short int key, int new_x, int new_y);
	void setAnimData(unsigned short int key, const Anim *, const Overlay *,
					 int af, int atz, bool ainvuln);
	void setFacing(unsigned short int key, MapDirection new_facing);
	void clear() { entities.clear(); } // delete all contained entities
	
	// getEntityGfx (used for drawing)
	// Coords of bottom-left of the map display area (square coord 0,0) must be passed in.
	// Entity depth is a base depth at which entities are drawn (it's modified by MapHeights).
	// Results will be _added_ to "gfx_buffer" (previous contents of "gfx_buffer" will be
	// retained as well).
	void getEntityGfx(int gvt, int bl_x, int bl_y, int entity_depth,
					  map<int, vector<GraphicElement> > &gfx_buffer);

private:
	// Commands are not executed immediately, but are stored up into a
	// "command block".  This way, if we get several commands over the
	// network in one go (due to lag or whatever), we can still spread
	// out their execution over several frames. (This can be thought
	// of as a form of interpolation.)
	struct Command {
		enum { MOVE, REPOSITION, SET_ANIM, SET_FACING } type;
		union {
			struct {
				// so: starting offset
				// nt: "natural time" (how long the move would take at normal speed).
				int so, nt;
				MotionType type;
			} move_info;
			struct {
				int x, y;
			} reposition_info;
			struct {
				// new appearance data, for an update command.
				const Anim *anim;
				const Overlay *ovr;
				int af;
				int atz;
				bool ainvuln;
			} anim_info;
			MapDirection facing_info;
		};
	};
	
	struct Data {
		int x, y;
		MapHeight height;
		const Anim *anim;
		const Overlay *ovr;
		int af;
		int atz;
		bool ainvuln;
		MapDirection facing;
		bool approached;          // true if currently "approached" (and not moving)
		
		int start_time;   // gvt at which the cmd block was started
		int finish_time;  // gvt at which the cmd block should be completed
		int tnt;          // total natural time (over all cmds)
		deque<Command> cmds;  // the command block itself
	};

	static int getFinalOffset(MotionType);
	void update(int, Data &dat);
	void update(int);
	void addGraphic(const Data &ent, int gvt, int bl_x, int bl_y,
					int entity_depth,
					map<int, vector<GraphicElement> > &gfx_buffer,
					int pixels_per_square, int points_per_square, int approach_offset);
	
	// key --> ID number for the entity;
	// data --> current client-side position, together with appearance
	// info (anim, overlay, facing etc) as well as "command block"
	// showing future changes to position and appearance that have not
	// yet been applied.
	map<unsigned short int, Data> entities;
};

#endif
