/*
 * lockable.cpp
 *
 * Copyright (c) Stephen Thompson, 2006.
 * Licenced for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "dungeon_map.hpp"
#include "dungeon_view.hpp"
#include "item.hpp"
#include "item_type.hpp"
#include "knight.hpp"
#include "lockable.hpp"
#include "misc/rng.hpp"
#include "trap.hpp"

//
// opening/closing: private functions
//

void Lockable::unlockAndOpen(DungeonMap &dmap, const MapCoord &mc, shared_ptr<Creature> cr,
							 bool open_traps)
{
	open(dmap, mc);
	lock = 0;
	closed = false;
	if (open_traps) disarmTraps(dmap, mc, cr);
	else activateTraps(dmap, mc, cr);
}

void Lockable::doOpen(DungeonMap &dmap, const MapCoord &mc, shared_ptr<Creature> cr,
					  bool open_traps)
{
	if (isLocked()) {
		bool ok = checkUnlock(cr);
		if (!ok) return;
	}
	unlockAndOpen(dmap, mc, cr, open_traps);
}

bool Lockable::checkUnlock(shared_ptr<Creature> cr) const
{
	if (!isLocked()) return true;
	
	// Only Knights have the ability to unlock doors
	shared_ptr<Knight> kt = dynamic_pointer_cast<Knight>(cr);
	if (kt) {
		for (int i=0; i<kt->getBackpackCount(); ++i) {
			int key = kt->getBackpackItem(i).getKey();
			if (key == this->lock) return true;
		}
		kt->getDungeonView().flashMessage("LOCKED");
	}
	return false;
}

//
// opening/closing: public functions
//

void Lockable::onActivate(DungeonMap &dmap, const MapCoord &mc, shared_ptr<Creature> cr,
						  bool open_traps)
{
	if (closed) {
		doOpen(dmap, mc, cr, open_traps);
	} else {
		close(dmap, mc);
		closed = true;
	}
	Tile::onActivate(dmap, mc, cr, open_traps);
}

void Lockable::onClose(DungeonMap &dmap, const MapCoord &mc, shared_ptr<Creature> cr)
{
	if (!closed) {
		close(dmap, mc);
		closed = true;
	}
	Tile::onClose(dmap, mc, cr);
}

void Lockable::onOpen(DungeonMap &dmap, const MapCoord &mc, shared_ptr<Creature> cr,
					  bool open_traps)
{
	if (closed) {
		doOpen(dmap, mc, cr, open_traps);
	}
	Tile::onOpen(dmap, mc, cr, open_traps);
}

void Lockable::onSpecialOpen(DungeonMap &dmap, const MapCoord &mc, shared_ptr<Creature> cr)
{
	if (closed) {
		unlockAndOpen(dmap, mc, cr, true);
	}
	Tile::onOpen(dmap, mc, cr, true);
}


//
// lock- and trap-related stuff
//

void Lockable::generateLock(int nkeys)
{
	if (lock < 0) {
		if (nkeys <= 0) {
			lock = 0;
		} else {
			if (g_rng.getBool(float(-lock)/100.0f)) {
				lock = g_rng.getInt(1, nkeys+1);
			} else {
				lock = 0;
			}
		}
	}
}

void Lockable::disarmTraps(DungeonMap &dmap, const MapCoord &mc, shared_ptr<Creature> cr)
{
	if (trap) {
		const ItemType *itype = trap->getTrapItem();
		if (itype) {
			shared_ptr<Item> dummy;
			const bool can_drop = CheckDropSquare(dmap, mc, *itype, dummy);
			if (can_drop) {
				dummy.reset(new Item(*trap->getTrapItem()));
				dmap.addItem(mc, dummy);
			}
		}
		trap = shared_ptr<Trap>();
	}
}

void Lockable::activateTraps(DungeonMap &dmap, const MapCoord &mc, shared_ptr<Creature> cr)
{
	if (trap) {
		trap->spring(dmap, mc, cr);
		trap = shared_ptr<Trap>();
	}
}

void Lockable::setTrap(DungeonMap &dmap, const MapCoord &mc, shared_ptr<Creature> cr,
					   shared_ptr<Trap> newtrap)
{
	if (trap) {
		activateTraps(dmap, mc, cr);
	}
	trap = newtrap;
}

void Lockable::onHit(DungeonMap &dmap, const MapCoord &mc, shared_ptr<Creature> cr)
{
	if (trap && trap->activateOnHit()) {
		activateTraps(dmap, mc, cr);
	}
	Tile::onHit(dmap, mc, cr);
}
