/*
 * missile.cpp
 *
 * Copyright (c) Stephen Thompson, 2006.
 * Licenced for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "cfg.hpp"
#include "creature.hpp"
#include "dungeon_map.hpp"
#include "item.hpp"
#include "item_type.hpp"
#include "mediator.hpp"
#include "missile.hpp"
#include "rng.hpp"
#include "task_manager.hpp"
#include "tile.hpp"

class MissileTask : public Task {
public:
	MissileTask(weak_ptr<Missile> m) : missile(m) { }
	virtual void execute(TaskManager &);
private:
	weak_ptr<Missile> missile;
};

void MissileTask::execute(TaskManager &tm)
{
	// Check missile is valid
	shared_ptr<Missile> m(missile.lock());
	if (!m) return;
	if (!m->getMap()) return;
	ASSERT(!m->isMoving());

	bool delete_from_map = false;
	
	// Check to see if we have hit something
	shared_ptr<Creature> cr = m->getMap()->getTargetCreature(*m, false);
	if (cr) {
		const int hit_chance = m->range_left * m->itype.getMissileHitMultiplier() + 1;
		if (g_rng.getBool(1.0f - 1.0f / hit_chance)) {
			const RandomInt *st = m->itype.getMissileStunTime();
			const RandomInt *dam = m->itype.getMissileDamage();
			const int stun_time = st? st->get() : 0;
			const int damage = dam? dam->get() : 0;
			if (damage > 0) {
				cr->damage(damage, stun_time + tm.getGVT());
			}
			delete_from_map = true;
		}
	}

	// Decrease range (taking stairs into account)
	MapCoord mc_ahead = DisplaceCoord(m->getPos(), m->getFacing());
	--(m->range_left);
	vector<shared_ptr<Tile> > tiles;
	m->getMap()->getTiles(mc_ahead, tiles);
	for (vector<shared_ptr<Tile> >::iterator it = tiles.begin(); it != tiles.end(); ++it) {
		if ((*it)->isStair()) {
			MapDirection dn = (*it)->stairsDownDirection();
			if (dn == m->getFacing()) {
				// going down. add to range
				if ((mc_ahead.getX() + mc_ahead.getY()) % 2 == 0) {
					++(m->range_left);
				}
			} else if (dn == Opposite(m->getFacing())) {
				// going up. subtract from range
				--(m->range_left);
			}
			break;
		}
	}
	if (m->range_left <= 0) delete_from_map = true;
	
	// Check access ahead.
	if (!delete_from_map) {
		MapAccess ma = m->getMap()->getAccess(mc_ahead,
			MapHeight(H_MISSILES + m->getFacing()));
		if (ma == A_APPROACH) {
			// 'partial' missile access
			int chance = m->itype.getMissileAccessChance();
			if (g_rng.getBool((100-chance) / 100.0f)) {
				delete_from_map = true;
			}
		} else if (ma == A_BLOCKED) {
			delete_from_map = true;
		}
	}
	
	if (delete_from_map) {
		// Remove the missile from the map. Drop item if necessary.
		if (m->drop_after) {
			shared_ptr<Item> item(new Item(m->itype));
			DropItem(item, *m->getMap(), m->getPos(), m->getFacing(), shared_ptr<Creature>());
		}
		m->rmFromMap();
	} else {
		// Move on, and reschedule task.
		m->move(MT_MOVE);
		tm.addTask(shared_from_this(), TP_NORMAL, m->getArrivalTime() + 1);
	}
}

bool CreateMissile(DungeonMap &dmap, const MapCoord &mc, MapDirection dir,
				   const ItemType &itype, bool drop_after, bool with_strength)
{
	// Check if we can place a missile here
	if (!dmap.canPlaceMissile(mc, dir)) return false;
		
	// Create the missile
	shared_ptr<Missile> m(new Missile(itype, drop_after));
	if (with_strength) m->doubleRange();
	m->setFacing(dir);
	m->addToMap(&dmap, mc);
	m->move(MT_MOVE);
	shared_ptr<MissileTask> mt(new MissileTask(m));
	Mediator::instance().getTaskManager().addTask(mt, TP_NORMAL, m->getArrivalTime() + 1);
	return true;
}

Missile::Missile(const ItemType &it, bool da)
	: itype(it), range_left(it.getMissileRange()), drop_after(da)
{
	setAnim(it.getMissileAnim());
	setSpeed(it.getMissileSpeed());
}

MapHeight Missile::getHeight() const
{
	return MapHeight(H_MISSILES + getFacing());
}

