/*
 * p_display.hpp
 *
 * Copyright (c) Stephen Thompson, 2006.
 * Licenced for non-commercial use only. See LICENCE.txt for details.
 *
 * Display: represents a window or display device. Contains operations for drawing on the display
 * (abstraction of the drawing operations of SDL or DirectDraw or whatever).
 *
 * Uses a pimpl so that the details of the graphics library being used are not exposed to the rest
 * of the program. To port to a new gfx library, only need to write a new p_display.cpp (maybe there
 * will be other p_*.cpp routines to do as well, eg graphics factory ...)
 *
 */

#ifndef P_DISPLAY_HPP
#define P_DISPLAY_HPP

#include "boost/shared_ptr.hpp"
using namespace boost;

class ColourChange;
class DisplayImpl;
class Graphic;

class Display {
public:
	// ctor (NB no dtor is needed, this can be placed in the display_impl dtor if necessary.)
	Display();

	// get display size
	int getWidth() const;
	int getHeight() const;
	
	// begin: clears the screen
	// end: flips the buffers.
	void begin();
	void end();

	// clipping: rectangle is BL inclusive and TR exclusive
	// (TODO) may be better to have a sentry object for clipping, because I always
	// forget to turn it off after using it :)
	void setClipRectangle(int bl_x, int bl_y, int tr_x, int tr_y);
	void clearClipRectangle();  // turns off clipping

	// text: draw text string with BL corner at a certain point
	void drawText(int bl_x, int bl_y, const char *str);
	int getTextWidth(const char *str);
	
	// graphics: draw a Graphic with BL corner at a certain point
	// (TODO) I don't think "darken" is used any more; could remove it some time. 
	void drawGraphic(int bl_x, int bl_y, const Graphic &gfx, bool darken = false);

	// this draws a graphic with some of the colours changed.
	void drawGraphicWithCC(int bl_x, int bl_y, const Graphic &, const ColourChange &);

	// pixel plotting: (r, g and b range from 0 to 255.)
	void beginPixels();  // locks the display so pixels can be plotted or read
	void drawPixel(int x, int y, int r, int g, int b);
	void readPixel(int x, int y, int &r, int &g, int &b);
	void endPixels();    // unlocks the display
	
private:
	shared_ptr<DisplayImpl> pimpl;
};

#endif
