/*
 * gapi_graphic.cpp
 *
 * Copyright (c) Stephen Thompson, 2006.
 * Licenced for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "app/gapi_app2d.hpp"
#include "binary_io/binary_io.hpp"
#include "gapi_graphic.hpp"
#include "misc/my_exceptions.hpp"
#include "rstream/rstream.hpp"

#include <ios>
using namespace std;

int GetWidth(const Graphic &g)
{
	return g.getWidth();
}

int GetHeight(const Graphic &g)
{
	return g.getHeight();
}

int GetID(const Graphic &g)
{
	return g.getID();
}

Graphic * CreateGraphic(const string &rname, int x, int y, int r, int g, int b)
{
	return new Graphic(rname, r, g, b, x, y);
}

void DeleteGraphic(Graphic *g)
{
	delete g;
}

void SerializeGraphic(ostream &str, Graphic &g, int id)
{
	BinWriteString(str, g.rname);
	BinWriteInt(str, g.hx);
	BinWriteInt(str, g.hy);
	BinWriteInt(str, g.r);
	BinWriteInt(str, g.g);
	BinWriteInt(str, g.b);
	g.id = id;
}

Graphic * CreateGraphicFromStream(istream &str)
{
	string rname = BinReadString(str);
	int hx = BinReadInt(str);
	int hy = BinReadInt(str);
	int r = BinReadInt(str);
	int g = BinReadInt(str);
	int b = BinReadInt(str);
	return CreateGraphic(rname, hx, hy, r, g, b);
}


//
// Graphic class
//

Graphic::Graphic(const string &rname_, int r_, int g_, int b_, int hx_, int hy_)
  : width(0), height(0), id(0),
    rname(rname_), r(r_), g(g_), b(b_), hx(hx_), hy(hy_),
    use_col_key(false), col_key(0)
{
	// Load a graphic from a .BMP file.
	// This code is based loosely on a similar routine found in SDL.

	// NOTE: I have here assumed that we will be running on little-endian machines only.

	if (!GapiApp2D::self()) throw GraphicLoadFailed("display not yet open");

	// Open stream
	RStream str(rname_);

	// Read in the BMP file header
	int magic1 = BinReadByte(str);
	int magic2 = BinReadByte(str);
	if (magic1 != 'B' || magic2 != 'M') {
		throw GraphicLoadFailed("BMP load failed");
	}
	BinReadUlong(str);  // bf_size 
	BinReadUword(str);  // bf_reserved_1
	BinReadUword(str);  // bf_reserved_2
	unsigned int bf_off_bits = BinReadUlong(str);

	// Read the BITMAPINFOHEADER.
	unsigned int bi_width = 0, bi_height = 0, bi_planes = 0, bi_bit_count = 0,
		bi_compression = 0, bi_size_image = 0, bi_xpels_per_metre = 0,
		bi_ypels_per_metre = 0, bi_clr_used = 0, bi_clr_important = 0;

	unsigned int bi_size = BinReadUlong(str);	
	if (bi_size == 12) {
		bi_width = BinReadUword(str);
		bi_height = BinReadUword(str);
		bi_planes = BinReadUword(str);
		bi_bit_count = BinReadUword(str);
		bi_compression = 0;  // BI_RGB
	} else {
		bi_width = BinReadUlong(str);
		bi_height = BinReadUlong(str);
		bi_planes = BinReadUword(str);
		bi_bit_count = BinReadUword(str);
		bi_compression = BinReadUlong(str);
		bi_size_image = BinReadUlong(str);
		bi_xpels_per_metre = BinReadUlong(str);
		bi_ypels_per_metre = BinReadUlong(str);
		bi_clr_used = BinReadUlong(str);
		bi_clr_important = BinReadUlong(str);
	}

	if (!str) throw GraphicLoadFailed("BMP load failed");

	// For simplicity, 1-, 4- and 8-bpp bitmaps are not supported.
	if (bi_bit_count <= 8) {
		throw GraphicLoadFailed("1-bit, 4-bit and 8-bit BMPs not supported");
	}

	// We do not support any BMP compression.
	unsigned int rmask = 0, gmask = 0, bmask = 0;
	switch (bi_compression) {
	case 0:  // BI_RGB
		// If there are no masks then use the defaults
		if (bf_off_bits == 14 + bi_size) {
			// Default values for the BMP format:
			switch (bi_bit_count) {
			case 15:
			case 16:
				rmask = 0x7c00;
				gmask = 0x03e0;
				bmask = 0x001f;
				break;
			case 24:
			case 32:
				rmask = 0xff0000;
				gmask = 0x00ff00;
				bmask = 0x0000ff;
				break;
			default:
				throw GraphicLoadFailed("BMP load failed");
			}
			break;
		}

		// Fall through

	case 3: // BI_BITFIELDS
		switch (bi_bit_count) {
		case 15:
		case 16:
		case 32:
			rmask = BinReadUlong(str);
			gmask = BinReadUlong(str);
			bmask = BinReadUlong(str);
			break;
		default:
			throw GraphicLoadFailed("BMP load failed");
		}
		break;

	default:
		throw GraphicLoadFailed("compressed BMP files not supported");
	}

	// Compute shift values
	if (rmask == 0 || gmask == 0 || bmask == 0) throw GraphicLoadFailed("BMP load failed");
	unsigned int rshift = 0, gshift = 0, bshift = 0;
	unsigned int tmp = 0;
	for (tmp = rmask; (tmp & 1) == 0; tmp >>= 1, ++rshift);
	for (tmp = gmask; (tmp & 1) == 0; tmp >>= 1, ++gshift);
	for (tmp = bmask; (tmp & 1) == 0; tmp >>= 1, ++bshift);

	width = bi_width;
	height = bi_height;
	pixel_data.reset(new unsigned short[width*height]);

	// Read the pixel data.
	str.seekg(bf_off_bits, ios_base::beg);
	if (!str) throw GraphicLoadFailed("BMP load failed");

	int pad=0;
	if (bi_bit_count == 32) {
		// Each pixel takes up four bytes; no padding needed.
		pad = 0;
	} else if (bi_bit_count == 24) {
		// Each pixel takes up three bytes; width must be padded to a multiple of 4 bytes
		pad = ((width*3)%4) ? 4-((width*3)%4) : 0;
	} else {
		// Each pixel takes up two bytes; width must be padded to a multiple of 4 bytes
		pad = (width%2) * 2;
	}

	unsigned int tmp2;
	for (int y=height-1; y>=0; --y) {
		for (int x=0; x<width; ++x) {
			if (bi_bit_count == 32) {
				tmp2 = BinReadUlong(str);
			} else if (bi_bit_count == 24) {
				unsigned int x = BinReadUbyte(str);
				unsigned int y = BinReadUbyte(str);
				unsigned int z = BinReadUbyte(str);
				tmp2 = x + 256U*y + 65536U*z;
			} else {
				tmp2 = BinReadUword(str);
			}
			// (We assume the screen is in 16-bpp format.)
			unsigned short dat = GapiApp2D::self()->rgbToScreen((tmp2 & rmask) >> rshift,
																(tmp2 & gmask) >> gshift,
																(tmp2 & bmask) >> bshift);
			pixel_data[x + y*width] = dat;
		}
		str.seekg(pad, ios_base::cur); // skip any padding bytes
	}

	if (!str) throw GraphicLoadFailed("BMP load failed");

	if (this->r != -1) {
		use_col_key = true;
		col_key = GapiApp2D::self()->rgbToScreen(this->r, this->g, this->b);
	} else {
		use_col_key = false;
	}
}

unsigned short int Graphic::getPixel(int x, int y)
{
	if (x < 0 || x >= width || y < 0 || y >= height) return 0;
	else return pixel_data[x + y*width];
}

void Graphic::blit(int sx1, int sy1, int sx2, int sy2, int dx, int dy,
				   unsigned short *dst_ptr, int dpy_width, const ColourChange *cc) const
{
	// NOTE: Coords in this routine are referred to top left corner
	
	const int w = sx2 - sx1;
	const int h = sy2 - sy1;
	
	// Copy pixels to the off-screen buffer.
	// Note: 16 bits per pixel is assumed

	for (int j=0; j<h; ++j) {
		for (int i=0; i<w; ++i) {
			unsigned short dat = pixel_data[i+sx1 + (j+sy1)*width];
			if (cc) {
				int ri, gi, bi, ro, go, bo;
				GapiApp2D::self()->screenToRgb(dat,ri,gi,bi);
				if (cc->lookup(ri,gi,bi,ro,go,bo)) {
					dat = GapiApp2D::self()->rgbToScreen(ro,go,bo);
				}
			}
			if (!use_col_key || dat != col_key) {
				*(dst_ptr + (i+dx) + (j+dy)*dpy_width) = dat;
			}
		}
	}
}
