/*
 * sdl_graphic.cpp
 *
 * Copyright (c) Stephen Thompson, 2006.
 * Licenced for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "binary_io/binary_io.hpp"
#include "cfg.hpp"
#include "my_exceptions.hpp"
#include "sdl_graphic.hpp"
#include "sdl_pixels.hpp"
#include "rstream/rstream_rwops.hpp"

// (TODO) see if changing rle_accel flag makes any difference :)
const bool use_rle_accel = true;
const unsigned int sdl_color_key_flag = SDL_SRCCOLORKEY | (use_rle_accel ? SDL_RLEACCEL : 0);

int GetWidth(const Graphic &g) {
	return g.getWidth();
}

int GetHeight(const Graphic &g) {
	return g.getHeight();
}

int GetID(const Graphic &g) {
	return g.getID();
}

Graphic::Graphic(const string &rname_, int r_, int g_, int b_, int hx_, int hy_)
	: surface(0), dark_surface(0), hx(hx_), hy(hy_), r(r_), g(g_), b(b_), id(0),
	  rname(rname_), need_reload(false)
{
	load();
}

Graphic::Graphic(const Graphic &original, const ColourChange &cc)
	: surface(0), dark_surface(0), hx(original.hx), hy(original.hy), r(original.r),
	  g(original.g), b(original.b), rname(original.rname), col_chg(cc), need_reload(false)
{
	if (original.surface) {
		// Make a copy of the original surface, changing the colours as necessary.
		int ro, go, bo, rn, gn, bn;
		SDL_Surface *temp = SDL_CreateRGBSurface(SDL_HWSURFACE | (r>=0? SDL_SRCCOLORKEY : 0),
												 original.surface->w, original.surface->h,
												 original.surface->format->BitsPerPixel,
												 original.surface->format->Rmask,
												 original.surface->format->Gmask,
												 original.surface->format->Bmask,
												 original.surface->format->Amask);
		SDL_LockSurface(original.surface);
		SDL_LockSurface(temp);
		for (int y=0; y<temp->h; ++y) {
			for (int x=0; x<temp->w; ++x) {
				ReadPixel(original.surface, x, y, ro, go, bo);
				if (col_chg.lookup(ro, go, bo, rn, gn, bn)) {
					PlotPixel(temp, x, y, rn, gn, bn);
				} else {
					PlotPixel(temp, x, y, ro, go, bo);
				}
			}
		}
		SDL_UnlockSurface(temp);
		SDL_UnlockSurface(original.surface);
		if (r >= 0) {
			SDL_SetColorKey(temp, sdl_color_key_flag, SDL_MapRGB(temp->format, r, g, b));
		}
		surface = SDL_DisplayFormat(temp);
		SDL_FreeSurface(temp);
	} else {
		// We lost the original surface. Reload from disk.
		load();
	}
}

Graphic::~Graphic()
{
	unload();
}

void Graphic::load() const
{
	if (!surface) {
		// (we assume the filetype is BMP for now -- perhaps support
		// SDL_image or similar later)
		SDL_Surface *temp = 0;
		try {
			temp = SDL_LoadBMP_RW(RWFromRStream(rname), true);
		} catch (RStreamError &rs) {
			// convert RStreamErrors into GraphicLoadFailed errors
			throw GraphicLoadFailed(rs.what());
		}
		if (temp) {
			if (r >= 0) {
				SDL_SetColorKey(temp, sdl_color_key_flag, SDL_MapRGB(temp->format, r, g, b));
			}
			surface = SDL_DisplayFormat(temp);
			SDL_FreeSurface(temp);
		}
		if (!surface) throw GraphicLoadFailed(string("failed to load graphic: ") + rname);

		// Do colour changes if necessary...
		if (!col_chg.empty()) {
			int ro, go, bo, rn, gn, bn;
			SDL_LockSurface(surface);
			for (int y=0; y<surface->h; ++y) {
				for (int x=0; x<surface->w; ++x) {
					ReadPixel(surface, x, y, ro, go, bo);
					if (col_chg.lookup(ro, go, bo, rn, gn, bn)) {
						PlotPixel(surface, x, y, rn, gn, bn);
					}
				}
			}
			SDL_UnlockSurface(surface);
		}
	}
}

void Graphic::unload() const
{
	if (surface) {
		if (dark_surface) {
			SDL_FreeSurface(dark_surface);
			dark_surface = 0;
		}
		SDL_FreeSurface(surface);
		surface = 0;
	}
}

void Graphic::makeDarkSurface() const
{
	const int DARKEN_ALPHA = CfgInt("darken_alpha");

	if (!surface) return;      // must load() first
	if (dark_surface) return;  // already have the dark_surface

	// Create dark_surface
	dark_surface = SDL_CreateRGBSurface(SDL_HWSURFACE | sdl_color_key_flag, surface->w, surface->h,
		surface->format->BitsPerPixel, surface->format->Rmask, surface->format->Gmask, 
		surface->format->Bmask, surface->format->Amask);
	if (!dark_surface) return;

	// Temporary surface
	SDL_Surface *temp = 0;

	try {
		if (r == -1) {
			// No colour key -- simple case

			// Clear dark_surface to black
			SDL_FillRect(dark_surface, 0, SDL_MapRGB(dark_surface->format, 0, 0, 0));

			// Blit our image onto dark_surface, using the colour key and DARKEN_ALPHA
			SDL_SetAlpha(surface, SDL_SRCALPHA, DARKEN_ALPHA);
			int blit_err = SDL_BlitSurface(surface, 0, dark_surface, 0);
			SDL_SetAlpha(surface, 0, 0);
			if (blit_err) throw UnexpectedError("blit failure during makeDarkSurface");

		} else {
			// Colour key -- have to go through some hoops in this case.
			// set up colour key & background colour on dark_surface
			SDL_SetColorKey(dark_surface, sdl_color_key_flag, SDL_MapRGB(dark_surface->format, r, g, b));
			SDL_FillRect(dark_surface, 0, SDL_MapRGB(dark_surface->format, r, g, b));
			
			// Create a temporary surface
			SDL_Surface *temp = SDL_CreateRGBSurface(SDL_SRCALPHA, surface->w, surface->h, 32, 
			0xff000000, 0xff0000, 0xff00, 0xff);
			if (!temp) return;
			
			// Fill the temporary surface with black
			SDL_FillRect(temp, 0, SDL_MapRGBA(temp->format, 0, 0, 0, 0));
			
			// Blit our image onto the temp surface, using the colour key and DARKEN_ALPHA
			// Dest alpha channel ends up being 255 (opaque) for pixels in the image, 0 otherwise.
			SDL_SetAlpha(surface, SDL_SRCALPHA, DARKEN_ALPHA);
			int blit_err = SDL_BlitSurface(surface, 0, temp, 0);
			SDL_SetAlpha(surface, 0, 0);
			if (blit_err) throw UnexpectedError("blit failure during makeDarkSurface");
			
			// Now alpha-blit the temp surface onto dark_surface, using the alpha mask; this copies the
			// darkened image pixels but leaves the colour key in the other pixels.
			SDL_SetAlpha(temp, SDL_SRCALPHA, 0);
			blit_err = SDL_BlitSurface(temp, 0, dark_surface, 0);
			if (blit_err) throw UnexpectedError("blit failure during makeDarkSurface");

			// Free the temporary surface
			SDL_FreeSurface(temp);
			temp = 0;
		}

		// ensure dark_surface is in the display format...
		temp = SDL_DisplayFormat(dark_surface);
		if (!temp) throw UnexpectedError("SDL error in makeDarkSurface");
		SDL_FreeSurface(dark_surface);
		dark_surface = temp;
		temp = 0;
		
	} catch (...) {
		// error -- delete everything (prevent memory leaks)
		if (temp) {
			SDL_FreeSurface(temp);
			temp = 0;
		}
		if (dark_surface) {
			SDL_FreeSurface(dark_surface);
			dark_surface = 0;
		}
		throw;
	}
}


void Graphic::blit(SDL_Surface *dest, int sdl_x, int sdl_y, bool darken) const
{
	// Try to reload if necessary...
	if (need_reload) {
		if (surface) {
			int result = SDL_LockSurface(surface);
			if (result != 0) return;  // can't yet reload
			SDL_UnlockSurface(surface);
			unload();
		}
		try {
			load();
			need_reload = false;
		} catch (GraphicLoadFailed&) {
			// do nothing (need_reload is still set so will try again next time.)
		}
	}
	
	// Check if the surface is loaded.
	if (!surface) return;

	// Make the dark surface if necessary.
	if (darken && !dark_surface) {
		makeDarkSurface();
		if (!dark_surface) return;  // surface not loaded
	}

	// correct for handle position
	sdl_x -= hx;
	sdl_y += hy;

	// Set up dest_rect (SDL ignores the width and height parts, only need to set x and y)
	SDL_Rect dest_rect;
	dest_rect.x = sdl_x;
	dest_rect.y = sdl_y;
	
	// Do the blit, and handle any errors.
	int result = SDL_BlitSurface(darken ? dark_surface : surface, 0, dest, &dest_rect);
	if (result == -1) {
		throw UnexpectedError("blit failure");
	} else if (result == -2) {
		// This means that the surface was lost.
		need_reload = true;
	}
}

Graphic * CreateGraphic(const string &filename, int x, int y, int r, int g, int b)
{
	return new Graphic(filename, r, g, b, x, y);
}

void DeleteGraphic(Graphic *g)
{
	delete g;
}


//
// (binary) serialization
//

void SerializeGraphic(ostream &str, Graphic &g, int id)
{
	BinWriteString(str, g.rname);
	BinWriteInt(str, g.hx);
	BinWriteInt(str, g.hy);
	BinWriteInt(str, g.r);
	BinWriteInt(str, g.g);
	BinWriteInt(str, g.b);
	if (!g.col_chg.empty()) {
		throw UnexpectedError("serialization of internal 'colour change' graphic requested");
	}
	g.id = id;
}

Graphic * CreateGraphicFromStream(istream &str)
{
	string rname = BinReadString(str);
	int hx = BinReadInt(str);
	int hy = BinReadInt(str);
	int r = BinReadInt(str);
	int g = BinReadInt(str);
	int b = BinReadInt(str);
	return CreateGraphic(rname, hx, hy, r, g, b);
}
