/*
 * potion_renderer.cpp
 *
 * Copyright (c) Stephen Thompson, 2006.
 * Licenced for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "binary_io/binary_io.hpp"
#include "cfg.hpp"
#include "client_config.hpp"
#include "platform/colour_change.hpp"
#include "platform/display.hpp"
#include "platform/graphic.hpp"
#include "potion_renderer.hpp"
#include "task_manager.hpp"

PotionRenderer::PotionRenderer()
	: def_r(-1), def_g(-1), def_b(-1)
{ }

void PotionRenderer::addGraphic(const Graphic *g)
{
	gfx.push_back(g);
}

void PotionRenderer::addColour(int r, int g, int b)
{
	if (r < 0 || g < 0 || b < 0 || r > 255 || g > 255 || b > 255) {
		return;
	}
	
	if (def_r < 0) {
		// First colour gets set as "default".
		def_r = r;
		def_g = g;
		def_b = b;
	} else {
		// Second and subsequent colours get added as ColourChanges.
		ColourChange cc;
		cc.add(def_r, def_g, def_b, r, g, b);
		colours.push_back(cc);
	}

}

void PotionRenderer::draw(int gvt, Display &disp, int health, PotionMagic magic,
						  bool poison_immunity, int left, int bottom) const
{
	disp.clearClipRectangle();

	// Find out which graphic to draw
	if (health < 0 || health >= gfx.size()) return;  // health out of range
	const Graphic *gr = gfx[health];

	// Check if poison immunity colour change should be used.
	if (poison_immunity) {
		const int s = (gvt / CfgInt("pot_flash_time")) % CfgInt("pi_flash_delay");
		if (s == 0) {
			// poison immunity: Use a white colour change
			ColourChange cc;
			cc.add(def_r, def_g, def_b, 255, 255, 255);
			disp.drawGraphicWithCC(left, bottom, *gr, cc);
			return;
		}
	}

	// Otherwise, work out colour to use based on "magic" value.
	if (magic < NO_POTION || magic > SUPER) return;  // magic out of range
	
	if (magic == NO_POTION) {
		// no potion: draw potion w/o colour change
		disp.drawGraphic(left, bottom, *gr);
	} else if (magic == SUPER) {
		// super: use a random colour change
		// Use gvt as a random seed (we should not disturb g_rng during rendering).
		int xx = (gvt/CfgInt("pot_flash_time"))*1103515245;
		int r = (xx >> 16) & 0xff;
		int g = (xx >> 8) & 0xff;
		int b = (xx & 0xff);
		ColourChange cc;
		cc.add(def_r, def_g, def_b, r, g, b);
		disp.drawGraphicWithCC(left, bottom, *gr, cc);
	} else {
		// draw with colour change appropriate to current magic
		if (magic-1 < 0 || magic-1 >= colours.size()) return;
		disp.drawGraphicWithCC(left, bottom, *gr, colours[int(magic-1)]);
	}
}

void PotionRenderer::serialize(ostream &str) const
{
	BinWriteLong(str, def_r);
	BinWriteLong(str, def_g);
	BinWriteLong(str, def_b);
	BinWriteLong(str, gfx.size());
	for (int i=0; i<gfx.size(); ++i) {
		BinWriteLong(str, GetID(*gfx[i]));
	}
	BinWriteLong(str, colours.size());
	for (int i=0; i<colours.size(); ++i) {
		colours[i].serialize(str);
	}
}

PotionRenderer::PotionRenderer(istream &str, ClientConfig &conf)
{
	def_r = BinReadLong(str);
	def_g = BinReadLong(str);
	def_b = BinReadLong(str);
	int sz = BinReadLong(str);
	gfx.reserve(sz);
	for (int i=0; i<sz; ++i) {
		gfx.push_back(conf.getGraphic(BinReadLong(str)));
	}
	sz = BinReadLong(str);
	colours.reserve(sz);
	for (int i=0; i<sz; ++i) {
		colours.push_back(ColourChange(str));
	}
}
