/*
 * remote_dungeon_view.hpp
 *
 * Copyright (c) Stephen Thompson, 2006.
 * Licenced for non-commercial use only. See LICENCE.txt for details.
 *
 * This serializes all DungeonView commands and transmits them through a network socket
 * (represented by an ostream).
 *
 */

#ifndef REMOTE_DUNGEON_VIEW_HPP
#define REMOTE_DUNGEON_VIEW_HPP

#include "dungeon_view.hpp"

#include "boost/multi_array.hpp"

#include <iosfwd>
using namespace std;

class ServerConfig;

class RemoteDungeonView : public DungeonView {
public:
	RemoteDungeonView(iostream &s, const ServerConfig &c)
		: str(s), cfg(c), current_room(-1), cur_room_width(0), cur_room_height(0),
		  control(0), game_over(false) { }
	
	virtual void setCurrentRoom(int r, int width, int height);
	virtual int getCurrentRoom() const;
	virtual void mapCurrentRoom(const MapCoord &bl_pos);
	virtual void mapWall(const MapCoord &mc);
	virtual void mapKnightLocation(int n, const MapCoord &mc);
	virtual void mapItemLocation(const MapCoord &mc, bool on);
	virtual void wipeMap();
	virtual bool isSquareMapped(const MapCoord &mc) const;
	
	virtual void addEntity(unsigned short int id, int x, int y, MapHeight ht, MapDirection facing,
						   const Anim *anim, const Overlay *ovr, int af, int atz_diff,
						   bool ainvuln,
						   int cur_ofs, MotionType mt, int motion_time_remaining);
	virtual void rmEntity(unsigned short int id);
	virtual void repositionEntity(unsigned short int id, int new_x, int new_y);
	virtual void moveEntity(unsigned short int id, MotionType motion_type,
							int motion_duration);
	virtual void setAnimData(unsigned short int id, const Anim *, const Overlay *, int af,
							 int atz_diff, bool ainvuln);
	virtual void setFacing(unsigned short int id, MapDirection new_facing);

	virtual void clearTiles(int x, int y);
	virtual void setTile(int x, int y, int depth, const Graphic *gfx,
						 shared_ptr<const ColourChange> cc, MiniMapColour col);
	virtual void setItem(int x, int y, const Graphic *g);
	virtual void placeIcon(int x, int y, const Graphic *g, int dur);

	virtual void setBackpack(int slot, const Graphic *g, int no_carried, int no_max);
	virtual void addSkull();
	virtual void setHealth(int h);
	virtual void setPotionMagic(PotionMagic pm, bool pi);

	virtual void flashMessage(const string &msg);
	virtual void cancelContinuousMessages();
	virtual void addContinuousMessage(const string &msg);
	virtual void flashScreen(int delay);
	
	virtual void winGame();
	virtual void loseGame();
	virtual void quitGame();
	
	virtual void setAvailableControls(const vector<pair<const Control*, bool> > &avail_ctrls);
	virtual void setMenuHighlight(const Control *ctrl);

	virtual const Control * readController();

	virtual void update();
	
private:
	// These functions handle sending stuff. The strategy for sending various things
	// (eg, the number of bits/bytes allocated) can be changed simply by changing the
	// following functions.
	void sendCommandNo(int x);  // this also sends an "update time" msg if needed....
	void sendRoomNo(int r)           { sendUbyte(r); }
	void sendMapCoord(const MapCoord &mc);
	void sendRoomCoord(int x, int y) { sendNibbles(x+1, y+1); }
	void sendSpeed(int spd) { sendUshort(spd); }
	void sendTileInfo(int d, MiniMapColour c, bool cc) { sendNibbles(d+8, (c<<1)+(cc?1:0)); }
	void sendAnimPtr(const Anim *);
	void sendColourChange(const ColourChange &);
	void sendControlPtr(const Control *);
	void sendControlPlusBool(const Control *, bool);
	void sendGraphicPtr(const Graphic *);
	void sendOverlayPtr(const Overlay *);
	void sendString(const string &);

	void sendNibbles(int, int);  // two unsigned 4-bit values
	void sendUbyte(int);
	void sendUshort(int);
	void sendLong(int);
	
private:
	iostream &str;
	const ServerConfig &cfg;
	int current_room, cur_room_width, cur_room_height;
	const Control *control;
	bool game_over;

	// The DungeonView is responsible for remembering which squares are mapped. (This is
	// so that Sense Items can work.)
	// We use a boost::multi_array, because this is easy to resize dynamically.
	// (I'm lazy and I don't want to have to pass in the map size to RemoteDungeonView at
	// construction time... so I just resize the array dynamically as needed!)
	boost::multi_array<bool, 2> mapped_squares;
};

#endif

