/*
 * server_config.hpp
 *
 * Copyright (c) Stephen Thompson, 2006.
 * Licenced for non-commercial use only. See LICENCE.txt for details.
 *
 * ServerConfig stores all the itemtypes, tiles, etc from the config files.
 * It acts as an interface between the server and the initialization code.
 *
 */

#ifndef SERVER_CONFIG_HPP
#define SERVER_CONFIG_HPP

#include "dungeon_layout.hpp"  // for BlockType
#include "item_type.hpp"       // for ItemSize
#include "map_support.hpp"
#include "menu.hpp"
#include "monster_type.hpp"
#include "platform/colour_change.hpp"
#include "potion_renderer.hpp"
#include "segment_set.hpp"
#include "skull_renderer.hpp"

#include "kconfig/kconfig_fwd.hpp"
#include "kconfig/kconfig_loader.hpp"

#include "boost/scoped_ptr.hpp"
#include "boost/shared_ptr.hpp"
using namespace boost;

#include <map>
using namespace std;

class Action;
class Anim;
class Control;
class GoreManager;
class Graphic;
class ItemGenerator;
class MenuInt;
class MonsterManager;
class Player;
class Tile;

class ServerConfig {
private:
	struct DungeonBlock {
		BlockType bt;
		string exits;
	};

	struct Sentry {
		Sentry(ServerConfig &c) : cfg(c) { }
		~Sentry();
		ServerConfig &cfg;
	};
	friend class Sentry;

	struct Popper {
		Popper(KConfig::KFile &k) : kf(k) { }
		~Popper();
		KConfig::KFile &kf;
	};

	struct SegmentTileData {
		vector<shared_ptr<Tile> > tiles;
		vector<const ItemType*> items;
	};
	
public:
	ServerConfig();
	~ServerConfig() { deleteAll(); }

	// Clear all data and load the given config file.
	// (the stream must have been opened in binary mode!)
	void loadConfigFile(const string &config_filename);

	//
	// Accessors
	//

	const SegmentSet & getSegmentSet() const { return segment_set; }
	shared_ptr<Tile> getWall() const { return wall; }
	shared_ptr<Tile> getHorizDoor(int i) const { return horiz_door[i]; }
	shared_ptr<Tile> getVertDoor(int i) const { return vert_door[i]; }

	const Menu & getMenu() const { return menu; }

	const ItemType * getDefaultItem() const { return default_item; }
	
	int getNumKnightAnims() const;
	const Anim * getKnightAnim(int i) const;
	shared_ptr<const ColourChange> getSecuredCC(int i) const;

	const vector<const Control *> & getControlSet() const { return control_set; }
	const vector<const Control *> & getCheatControls() const { return cheat_controls; }
	const Control * getControlFromID(int n) const;

	const Graphic * getStuffBagGraphic() const { return stuff_bag_graphic; }
	
	// skull, potion renderers
	const SkullRenderer &getSkullRenderer() const { return skull_renderer; }
	const PotionRenderer &getPotionRenderer() const { return potion_renderer; }

	// set up monsters and gore manager:
	void configureGoreManager(GoreManager &gm, const vector<Player*> &players) const;
	void configureMonsterManager(MonsterManager &mm, int vampire_bats,
								 int zombie_activity) const;
	
	// in-game menu:
	const Graphic * getMenuCentre() const { return menu_centre; }
	const Graphic * getMenuEmpty() const { return menu_empty; }
	const Graphic * getMenuHighlight() const { return menu_highlight; }
	
	
	//
	// Interface to KFile. Made public, because used by ActionMakers etc.
	//

	KConfig::KFile * getKFile() { return kf.get(); }   // NULL if file not opened.

	MapAccess popAccessCode(MapAccess dflt);
	void popAccessTable(MapAccess out[], MapAccess dflt);
	Action * popAction();
	Action * popAction(Action *dflt);
	Anim * popAnim(bool batmode = false);
	Anim * popAnim(Anim *dflt, bool batmode = false);
	BlockType popBlockType();
	bool popBool();
	bool popBool(bool dflt);
	Control * popControl();
	Control * popControl(Control *dflt);
	void popControlSet(vector<const Control*> &which_control_set);
	DungeonBlock popDungeonBlock();
	DungeonDirective * popDungeonDirective();
	DungeonLayout * popDungeonLayout();
	Graphic * popGraphic();
	Graphic * popGraphic(Graphic *dflt);
	void popHouseColours();
	ItemGenerator * popItemGenerator();
	ItemSize popItemSize();
	ItemSize popItemSize(ItemSize dflt);
	ItemType * popItemType();
	ItemType * popItemType(ItemType *dflt);
	MapDirection popMapDirection();
	MapDirection popMapDirection(MapDirection dflt);
	void popMenu();
	MenuItem * popMenuItem();  // returns id string & menu item
	MenuInt * popMenuInt();
	MenuValue * popMenuValue();
	void popMenuValueName(MenuValue &);
	void popMenuValueDirective(MenuValue &);
	Overlay * popOverlay();
	Overlay * popOverlay(Overlay *dflt);
	void popOverlayOffsets(int *);
	void popPotionRenderer();
	int popProbability();  // pop a float, multiply by 100 and round to nearest int. (Error if result <0 or >100.)
	int popProbability(int dflt);
	RandomDungeonLayout * popRandomDungeonLayout();
	void popRGB(int &r, int &g, int &b);  // returned values are in range 0-255.
	Segment * popSegment(const vector<SegmentTileData> &tile_map, int &category);
	Segment * popSegmentData(const vector<SegmentTileData> &tile_map, int w, int h);
	void popSegmentList(const vector<SegmentTileData> &tile_map);
	void popSegmentRooms(Segment &);
	void popSegmentSet();
	void popSegmentSwitches(Segment &);
	void popSegmentTiles(vector<SegmentTileData> &tile_map);
	void popSegmentTile(SegmentTileData &seg_tile, bool &items_yet);
	void popSkullRenderer();
	struct StairInfo {
		bool is_stair;
		bool special;
		MapDirection down_direction;
	};
	StairInfo popStairsDown();
	shared_ptr<Tile> popTile();
	shared_ptr<Tile> popTile(shared_ptr<Tile> dflt);
	int popTileItems(const MapAccess acc[]);  // acc is just there to provide a default answer.
	void popTileList(vector<shared_ptr<Tile> > &output);
	int popTime();  // pop a float, convert it from seconds to ticks. (Error if result negative.)
	int popTime(int dflt);

	int getSegmentCategory(const string &); // returns -1 if empty string given
	int getTileCategory(const string &); // ditto
	void useSegmentCategory(int sc) { doUse(sc, segment_categories.size(), segment_categories_used); }
	void useTileCategory(int tc) { doUse(tc, tile_categories.size(), tile_categories_used); }
	
	KConfig::RandomIntContainer & getRandomIntContainer() { return random_ints; }
	

	//
	// serialization interface
	// This serializes Graphics, Anims, Overlays, Controls (ID numbers are also assigned)
	// as well as PotionRenderer, SkullRenderer, and key Menu graphics.
	//
	void serializeAll(ostream &);
	
	
private:
	void deleteAll();
	DungeonLayout * doPopDungeonLayout(int w, int h);
	void defineSegmentCategory(int sc) { doUse(sc, segment_categories.size(), segment_categories_defined); }
	void defineTileCategory(int tc) { doUse(tc, tile_categories.size(), tile_categories_defined); }
	void doUse(int item, int sz, vector<bool> &container);
	void processTile(const vector<SegmentTileData> &tmap, Segment &seg, int x, int y);
	const SegmentTileData & tileFromInt(const vector<SegmentTileData> &tile_map, int t);
	
private:
	// Storage for all 'basic' game objects, ie stuff loaded directly from the config file.
	// These last for the lifetime of the program basically. (Deleted only by ~ServerConfig.)
	// (TODO) would be better to use some sort of hash tables rather than maps here.
	map<const Value *, Action *> actions;
	map<const Value *, Anim *> anims;
	map<const Value *, Control *> controls;
	vector<Control*> controls_by_id;
	map<const Value *, DungeonDirective *> dungeon_directives;
	map<const Value *, DungeonLayout *> dungeon_layouts;
	map<const Value *, Graphic *> graphics;
	map<const Value *, ItemGenerator *> item_generators;
	map<const Value *, ItemType *> item_types;
	vector<ItemType*> special_item_types;
	map<const Value *, MenuItem *> menu_items;
	map<const Value *, MenuInt *> menu_ints;
	map<const Value *, MenuValue *> menu_values;
	map<const Value *, Overlay *> overlays;
	map<const Value *, RandomDungeonLayout *> random_dungeon_layouts;
	KConfig::RandomIntContainer random_ints;
	map<const Value *, Segment *> segments;
	map<const Value *, shared_ptr<Tile> > tiles;
	SegmentSet segment_set;
	Menu menu;

	// various stuff ...
	shared_ptr<Tile> wall, horiz_door[2], vert_door[2];
	vector<ColourChange> house_colours_normal;  // one entry per House
	vector<ColourChange> house_colours_invulnerable;
	vector<shared_ptr<const ColourChange> > secured_cc;
	Anim * knight_anim;  // In colours of house 0, ie house_colours_xxx[0] is always 'empty'.
	mutable vector<Anim*> knight_anims;  // copies of "knight_anim" with approp. col-changes
	ItemType * default_item;
	vector<const Control*> control_set, cheat_controls;
	SkullRenderer skull_renderer;
	PotionRenderer potion_renderer;
	const Graphic * menu_centre;
	const Graphic * menu_empty;
	const Graphic * menu_highlight;
	const Graphic * stuff_bag_graphic;
	
	// monsters, gore
	const Graphic *blood_icon;
	vector<shared_ptr<Tile> > blood_tiles, dead_knight_tiles, dead_bat_tiles;
	shared_ptr<Tile> dead_zombie_tile;
	vector<shared_ptr<Tile> > other_zombie_tiles;
	vector<shared_ptr<Tile> > bat_pit_tiles;
	scoped_ptr<MonsterType> vampire_bat_type, zombie_type;
	int bat_speed, zombie_speed;
	const KConfig::RandomInt * bat_health;
	const KConfig::RandomInt * zombie_health;
	const Anim * bat_anim;
	const Anim * zombie_anim;
	const ItemType * zombie_weapon;
	vector<shared_ptr<Tile> > zombie_ai_avoid;
	const ItemType * zombie_ai_hit;
	const ItemType * zombie_ai_fear;
	
	
	// Some maps from names (in the config system) to integer id codes.
	// These are non-empty only during loadConfigFile.
	map<string, int> menu_item_names;
	map<string, int> segment_categories;
	map<string, int> tile_categories;
	vector<bool> segment_categories_used;
	vector<bool> segment_categories_defined;
	vector<bool> tile_categories_used;
	vector<bool> tile_categories_defined;

	// kf is non-null only during loadConfigFile.
	shared_ptr<KConfig::KFile> kf;
};

#endif
