/*
 * tile.cpp
 *
 * Copyright (c) Stephen Thompson, 2006.
 * Licenced for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "action.hpp"
#include "dungeon_map.hpp"
#include "item.hpp"
#include "item_type.hpp"
#include "mediator.hpp"
#include "sweep.hpp"
#include "tile.hpp"

//
// public functions
//

Tile::Tile()
	: graphic(0), depth(0), is_stair(false), items_mode(BLOCKED), 
	  stair_direction(D_NORTH), hit_points(0), initial_hit_points(0), on_activate(0), 
	  on_close(0), on_open(0), on_walk_over(0), on_approach(0), on_withdraw(0), on_hit(0),
	  control(0)
{
	for (int i=0; i<=H_MISSILES; ++i) access[i] = A_CLEAR;
}

void Tile::construct(const Graphic *g, int d,
					 bool ita, bool dstry, int item_category_, 
					 const MapAccess acc[],
					 bool is, bool stop, MapDirection sdir, const RandomInt * ihp,
					 shared_ptr<Tile> dto,
					 const Action *ac, const Action *c, const Action *o,
					 const Action *wo, const Action *ap, const Action *wi, const Action *hit,
					 const Control *ctrl)
{
	graphic = g; depth=d; item_category = item_category_;
	is_stair=is; stair_top=stop; stair_direction=sdir; 
	initial_hit_points=ihp; destroy_to = dto;
	on_activate = ac; on_close = c; on_open = o;
	on_walk_over = wo; on_approach = ap; on_withdraw = wi;
	on_hit = hit; control = ctrl;
	for (int i=0; i<=H_MISSILES; ++i) { access[i] = acc[i]; }

	items_mode = BLOCKED;
	if (ita) items_mode = ALLOWED;
	if (dstry) items_mode = DESTROY;
}

void Tile::construct(const Graphic *g, int dpth)
{
	graphic = g; depth=dpth;
	items_mode = ALLOWED;
	// the rest take default values as set in the ctor.
}

void Tile::construct(const Action *walkover, const Action *activate)
{
	on_walk_over = walkover;
	on_activate = activate;
	items_mode = ALLOWED;
}


shared_ptr<Tile> Tile::clone()
{
	shared_ptr<Tile> new_tile = doClone();
	new_tile->setHitPoints();
	return new_tile;
}

void Tile::setHitPoints()
{
	if (initial_hit_points) {
		hit_points = initial_hit_points->get();
	}
}

void Tile::damage(DungeonMap &dmap, const MapCoord &mc, int amount)
{
	// NB: hit_points <= 0 means that the tile is indestructible.
	if (hit_points > 0) {
		hit_points -= amount;
		if (hit_points <= 0) {
			// Tile has been destroyed
			// (TODO) will probably have to replace destroy_to with on_destroy at some stage 
			// (for sound effects)?
			shared_ptr<Tile> self(shared_from_this());
			dmap.rmTile(mc, self);
			if (destroy_to) dmap.addTile(mc, destroy_to->clone());
			self->onDestroy(dmap, mc);
		}
	}
}

void Tile::onDestroy(DungeonMap &dmap, const MapCoord &mc)
{
	shared_ptr<Item> item = dmap.getItem(mc);
	if (item && item->getType().isFragile()) dmap.rmItem(mc);
}

void Tile::onActivate(DungeonMap &dmap, const MapCoord &mc, shared_ptr<Creature> actor,
					  bool open_traps)
{
	ActionData ad;
	ad.setActor(actor, false);
	ad.setTile(&dmap, mc, shared_from_this());
	ad.setFlag(open_traps);
	if (on_activate) {
		on_activate->execute(ad);
	} else if (on_close) {
		on_close->execute(ad);
	} else if (on_open) {
		on_open->execute(ad);
	}
}

void Tile::onClose(DungeonMap &dmap, const MapCoord &mc, shared_ptr<Creature> actor)
{
	if (on_close) {
		ActionData ad;
		ad.setActor(actor, false);
		ad.setTile(&dmap, mc, shared_from_this());
		on_close->execute(ad);
	}
}

void Tile::onOpen(DungeonMap &dmap, const MapCoord &mc, shared_ptr<Creature> actor,
				  bool open_traps)
{
	if (on_open) {
		ActionData ad;
		ad.setActor(actor, false);
		ad.setTile(&dmap, mc, shared_from_this());
		ad.setFlag(open_traps);
		on_open->execute(ad);
	}
}

void Tile::onWalkOver(DungeonMap &dmap, const MapCoord &mc, shared_ptr<Creature> actor)
{
	if (on_walk_over) {
		ActionData ad;
		ad.setActor(actor, false);
		ad.setTile(&dmap, mc, shared_from_this());
		on_walk_over->execute(ad);
	}
}

void Tile::onApproach(DungeonMap &dmap, const MapCoord &mc, shared_ptr<Creature> actor)
{
	if (on_approach) {
		ActionData ad;
		ad.setActor(actor, false);
		ad.setTile(&dmap, mc, shared_from_this());
		on_approach->execute(ad);
	}
}

void Tile::onWithdraw(DungeonMap &dmap, const MapCoord &mc, shared_ptr<Creature> actor)
{
	if (on_withdraw) {
		ActionData ad;
		ad.setActor(actor, false);
		ad.setTile(&dmap, mc, shared_from_this());
		on_withdraw->execute(ad);
	}
}

void Tile::onHit(DungeonMap &dmap, const MapCoord &mc, shared_ptr<Creature> actor)
{
	if (on_hit) {
		ActionData ad;
		ad.setActor(actor, false);
		ad.setTile(&dmap, mc, shared_from_this());
		on_hit->execute(ad);
	}
}



//
// protected functions
//

shared_ptr<Tile> Tile::doClone()
{
	ASSERT(typeid(*this)==typeid(Tile)); //doClone must be overridden in subclasses.

	// basic tiles are shared rather than cloned
	// however, if a tile has hit points, it must be copied. (and hit points must be rerolled, also.)
	if (initial_hit_points) {
		return shared_ptr<Tile>(new Tile(*this));
	} else {
		return shared_from_this();
	}
}

void Tile::setGraphic(DungeonMap *dmap, const MapCoord &mc, const Graphic *g,
					  shared_ptr<const ColourChange> ccnew)
{
	graphic = g;
	cc = ccnew;
	if (dmap) {
		// Graphic-change needs to be propagated to clients
		Mediator::instance().onChangeTile(*dmap, mc, *this);
	}
}

void Tile::setItemsAllowed(DungeonMap *dmap, const MapCoord &mc, bool allow, bool destroy) 
{
	if (destroy) items_mode = DESTROY;
	else if (allow) items_mode = ALLOWED;
	else items_mode = BLOCKED;
	if (dmap && items_mode != ALLOWED) {
		// We don't need to tell Mediator about items-allowed changes.
		// However we do need to call SweepItems if items-allowed is now BLOCKED, or destroy
		// item if mode is DESTROY.
		if (items_mode == DESTROY) {
			dmap->rmItem(mc);
		} else {
			SweepItems(*dmap, mc);
		}
	}
}

void Tile::setAccess(DungeonMap *dmap, const MapCoord &mc, MapHeight height, MapAccess acc)
{
	access[height] = acc;
	if (dmap) {
		// We don't need to tell Mediator about access changes, but we do need 
		// to call SweepCreatures.
		SweepCreatures(*dmap, mc, true, height);
	}
}

void Tile::setAccess(DungeonMap *dmap, const MapCoord &mc, MapAccess acc)
{
	for (int i=0; i<=H_MISSILES; ++i) {
		access[i] = acc;
	}
	SweepCreatures(*dmap, mc, false, H_MISSILES);
}


MiniMapColour Tile::getColour() const
{
	// Current method of determining mini map colour is as follows: colour is
	// COL_WALL if access level is A_BLOCKED at all available heights; else colour is
	// COL_FLOOR.
	// This can be overridden by subclasses if required.
	for (int h=0; h<=H_MISSILES; ++h) {
		if (getAccess(MapHeight(h)) != A_BLOCKED) {
			return COL_FLOOR;
		}
	}
	return COL_WALL;
}
